(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Option_Monad
imports
  Main
  "~~/src/HOL/Library/Monad_Syntax"
begin

lemma bind_Some_ex_conv[simp]:
  "m \<guillemotright>= f = Some y \<longleftrightarrow> (\<exists>x. m = Some x \<and> f x = Some y)"
  by (cases m) simp_all

lemma bind_cong[fundef_cong]:
  assumes "m1 = m2"
    and "\<And>x. m2 = Some x \<Longrightarrow> f1 x = f2 x"
  shows "m1 \<guillemotright>= f1 = m2 \<guillemotright>= f2"
  using assms by (cases m1) auto

definition guard :: "bool \<Rightarrow> unit option" where
  "guard b \<equiv> if b then Some () else None"

lemma guard_cong[fundef_cong]:
  "b = c \<Longrightarrow> (c \<Longrightarrow> m = n) \<Longrightarrow> guard b \<guillemotright> m = guard c \<guillemotright> n"
  by (simp add: guard_def)

lemma guard_simps:
  "guard b = Some x \<longleftrightarrow> b"
  "guard b = None \<longleftrightarrow> \<not> b"
  by (cases b) (simp_all add: guard_def)

lemma guard_elims[elim]:
  "guard b = Some x \<Longrightarrow> (b \<Longrightarrow> P) \<Longrightarrow> P"
  "guard b = None \<Longrightarrow> (\<not> b \<Longrightarrow> P) \<Longrightarrow> P"
  by (simp_all add: guard_simps)

lemma guard_intros[intro, simp]:
  "b \<Longrightarrow> guard b = Some ()"
  "\<not> b \<Longrightarrow> guard b = None"
  by (simp_all add: guard_simps)

lemma guard_True[simp]: "guard True = Some ()" by simp
lemma guard_False[simp]: "guard False = None" by simp

fun zip_option :: "'a list => 'b list => ('a \<times> 'b) list option" where
  "zip_option [] [] = Some []"
| "zip_option (x#xs) (y#ys) = do { zs \<leftarrow> zip_option xs ys; Some ((x, y) # zs) }"
| "zip_option (x#xs) [] = None"
| "zip_option [] (y#ys) = None"

text {* induction scheme for zip *}
lemma zip_induct[case_names Cons_Cons Nil1 Nil2]:
  assumes "\<And>x xs y ys. P xs ys \<Longrightarrow> P (x # xs) (y # ys)"
    and "\<And>ys. P [] ys"
    and "\<And>xs. P xs []"
  shows "P xs ys"
proof (induct xs arbitrary: ys)
  case Nil thus ?case by fact
next
  case (Cons x xs)
  note IH = this
  show ?case
  proof (cases ys)
    case Nil show ?thesis unfolding Nil by fact
  next
    case (Cons z zs)
    show ?thesis unfolding Cons using assms(1)[OF IH] .
  qed
qed

lemma zip_option_zip_conv:
  "zip_option xs ys = Some zs
    \<longleftrightarrow> length ys = length xs \<and> length zs = length xs \<and> zs = zip xs ys"
proof -
  {
    assume "zip_option xs ys = Some zs"
    hence "length ys = length xs \<and> length zs = length xs \<and> zs = zip xs ys"
    proof (induct xs ys arbitrary: zs rule: zip_option.induct)
      case (2 x xs y ys)
      then obtain zs' where "zip_option xs ys = Some zs'"
        and "zs = (x, y) # zs'" by (cases "zip_option xs ys") auto
      from 2(1)[OF this(1)] and this(2) show ?case by simp
    qed simp_all
  } moreover {
    assume "length ys = length xs" and "zs = zip xs ys"
    hence "zip_option xs ys = Some zs"
      by (induct xs ys arbitrary: zs rule: zip_induct) force+
  }
  ultimately show ?thesis by blast
qed

lemma zip_option_None:
  "zip_option xs ys = None \<longleftrightarrow> length xs \<noteq> length ys"
proof -
  {
    assume "zip_option xs ys = None"
    have "length xs \<noteq> length ys"
    proof (rule ccontr)
      assume "\<not> length xs \<noteq> length ys"
      hence "length xs = length ys" by simp
      hence "zip_option xs ys = Some (zip xs ys)" by (simp add: zip_option_zip_conv)
      with `zip_option xs ys = None` show False by simp
    qed
  } moreover {
    assume "length xs \<noteq> length ys"
    hence "zip_option xs ys = None"
      by (induct xs ys rule: zip_option.induct) simp_all
  }
  ultimately show ?thesis by blast
qed

declare zip_option.simps[simp del]

lemma zip_option_intros[intro]:
  "\<lbrakk>length ys = length xs; length zs = length xs; zs = zip xs ys\<rbrakk>
    \<Longrightarrow> zip_option xs ys = Some zs"
  "length xs \<noteq> length ys \<Longrightarrow> zip_option xs ys = None"
  by (simp_all add: zip_option_zip_conv zip_option_None)

lemma zip_option_elims[elim]:
  "zip_option xs ys = Some zs
    \<Longrightarrow> (\<lbrakk>length ys = length xs; length zs = length xs; zs = zip xs ys\<rbrakk> \<Longrightarrow> P)
    \<Longrightarrow> P"
  "zip_option xs ys = None \<Longrightarrow> (length xs \<noteq> length ys \<Longrightarrow> P) \<Longrightarrow> P"
  using assms by (simp_all add: zip_option_zip_conv zip_option_None)

lemma zip_option_simps[simp]:
  "zip_option xs ys = None \<Longrightarrow> length xs = length ys \<Longrightarrow> False"
  "zip_option xs ys = None \<Longrightarrow> length xs \<noteq> length ys"
  "zip_option xs ys = Some zs \<Longrightarrow> zs = zip xs ys"
(* leads to nontermination in Semantic_Labeling_Carrier
  "zip_option xs ys = Some zs \<Longrightarrow> length ys = length xs"
  "zip_option xs ys = Some zs \<Longrightarrow> length zs = length xs"*)
  by (simp_all add: zip_option_None zip_option_zip_conv)

end