(*
  Abstract syntax operations on IsaFoR terms.
*)

signature ISAFOR_TERM =
sig
  val termT : typ
  val substT : typ

  val var_const : term
  val fun_const : term

  val dummy_term : term
  val mk_var : string -> term
  val mk_app : string -> term list -> term
  val mk_term : term (*hol*) -> term (*isafor*)
  val mk_subst : (term * term) list -> term

  val dest_var : term -> string
  val dest_app : term -> string * term list

  val fold_term : (string -> 'a) -> (string -> 'a list -> 'a) -> (term -> 'a) -> term -> 'a
  val vars_of : term -> string list
end

structure Isafor_Term : ISAFOR_TERM =
struct


(* types and constants *)

val stringT = @{typ String.string}
val termT = @{typ "(String.string, String.string) Term.term"}
val substT = @{typ "(String.string, String.string) Term.subst"}

val var_const = Const (@{const_name Term.Var}, stringT --> termT)
val fun_const = Const (@{const_name Term.Fun},
  stringT --> HOLogic.listT termT --> termT)


(* construction *)

fun mk_var name = var_const $ HOLogic.mk_string name
fun mk_app fname ts = fun_const $ HOLogic.mk_string fname
  $ HOLogic.mk_list termT ts

fun mk_term (Var ((v,_), _)) = mk_var v (* FIXME: Var/Free may clash *)
  | mk_term (Free (v, _)) = mk_var v
  | mk_term t =
  let val (Const (c, _), ts) = strip_comb t
  in mk_app c (map mk_term ts) end

fun fun_upd_const T T' = Const (@{const_name Fun.fun_upd}, 
  (T --> T') --> T --> T' --> T --> T')
(*fun undefined_const T = Const (@{const_name "HOL.undefined"}, T)*)

fun mk_subst ps =
let fun subst [] = var_const
      | subst ((v,t)::ps) = 
        fun_upd_const HOLogic.stringT termT $ subst ps $ v $ t
in
  subst ps
end

val dummy_term = mk_app "dummy_term" []


(* destruction *)

fun dest_var (Const (@{const_name Term.Var}, _) $ n) =
  HOLogic.dest_string n
  | dest_var t = raise TERM("dest_var", [t])

fun dest_app (Const (@{const_name Term.Fun}, _) $ n $ ts) =
  (HOLogic.dest_string n, HOLogic.dest_list ts)
  | dest_app t = raise TERM("dest_app", [t])


(* traversal *)

fun fold_term v _ _ (Const (@{const_name Term.Var}, _) $ n) 
  = v (HOLogic.dest_string n)
  | fold_term v f ng (Const (@{const_name Term.Fun}, _) $ n $ ts)
  = f (HOLogic.dest_string n) (map (fold_term v f ng) (HOLogic.dest_list ts))
  | fold_term _ _ ng t = ng t

val vars_of = fold_term
  single 
  (fn _ => fn vss => fold (fold (insert (op =))) vss [])
  (fn t => raise TERM("vars_of", [t]))

end
